# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import PathScripts.PathLog as PathLog
import PathScripts.PathPreferences as PathPreferences
import PathScripts.PathPropertyBag as PathPropertyBag
import PathScripts.PathUtil as PathUtil
import PySide
import json
import os
import zipfile

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader
Part = LazyLoader('Part', globals(), 'Part')

__title__ = "Tool bits."
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecadweb.org"
__doc__ = "Class to deal with and represent a tool bit."

PropertyGroupShape = 'Shape'

_DebugFindTool = False

PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())
#PathLog.trackModule()


def translate(context, text, disambig=None):
    return PySide.QtCore.QCoreApplication.translate(context, text, disambig)

def _findToolFile(name, containerFile, typ):
    PathLog.track(name)
    if os.path.exists(name):  # absolute reference
        return name

    if containerFile:
        rootPath = os.path.dirname(os.path.dirname(containerFile))
        paths = [os.path.join(rootPath, typ)]
    else:
        paths = []
    paths.extend(PathPreferences.searchPathsTool(typ))

    def _findFile(path, name):
        PathLog.track(path, name)
        fullPath = os.path.join(path, name)
        if os.path.exists(fullPath):
            return (True, fullPath)
        for root, ds, fs in os.walk(path):
            for d in ds:
                found, fullPath = _findFile(d, name)
                if found:
                    return (True, fullPath)
        return (False, None)


    for p in paths:
        found, path = _findFile(p, name)
        if found:
            return path
    return None


def findToolShape(name, path=None):
    '''findToolShape(name, path) ... search for name, if relative path look in path'''
    PathLog.track(name, path)
    return _findToolFile(name, path, 'Shape')


def findToolBit(name, path=None):
    '''findToolBit(name, path) ... search for name, if relative path look in path'''
    PathLog.track(name, path)
    if name.endswith('.fctb'):
        return _findToolFile(name, path, 'Bit')
    return _findToolFile("{}.fctb".format(name), path, 'Bit')


def findToolLibrary(name, path=None):
    '''findToolLibrary(name, path) ... search for name, if relative path look in path'''
    PathLog.track(name, path)
    if name.endswith('.fctl'):
        return _findToolFile(name, path, 'Library')
    return _findToolFile("{}.fctl".format(name), path, 'Library')


def _findRelativePath(path, typ):
    PathLog.track(path, typ)
    relative = path
    for p in PathPreferences.searchPathsTool(typ):
        if path.startswith(p):
            p = path[len(p):]
            if os.path.sep == p[0]:
                p = p[1:]
            if len(p) < len(relative):
                relative = p
    return relative


def findRelativePathShape(path):
    return _findRelativePath(path, 'Shape')


def findRelativePathTool(path):
    return _findRelativePath(path, 'Bit')


def findRelativePathLibrary(path):
    return _findRelativePath(path, 'Library')

class ToolBit(object):

    def __init__(self, obj, shapeFile, path=None):
        PathLog.track(obj.Label, shapeFile, path)
        self.obj = obj
        obj.addProperty('App::PropertyFile', 'BitShape', 'Base', translate('PathToolBit', 'Shape for bit shape'))
        obj.addProperty('App::PropertyLink', 'BitBody', 'Base', translate('PathToolBit', 'The parametrized body representing the tool bit'))
        obj.addProperty('App::PropertyFile', 'File', 'Base', translate('PathToolBit', 'The file of the tool'))
        obj.addProperty('App::PropertyString', 'ShapeName', 'Base', translate('PathToolBit', 'The name of the shape file'))
        obj.addProperty('App::PropertyStringList', 'BitPropertyNames', 'Base', translate('PathToolBit', 'List of all properties inherited from the bit'))

        if path:
            obj.File = path
        if shapeFile is None:
            obj.BitShape = 'endmill.fcstd'
            self._setupBitShape(obj)
            self.unloadBitBody(obj)
        else:
            obj.BitShape = shapeFile
            self._setupBitShape(obj)
        self.onDocumentRestored(obj)

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        for obj in FreeCAD.ActiveDocument.Objects:
            if hasattr(obj, 'Proxy') and obj.Proxy == self:
                self.obj = obj
                break
        return None

    def onDocumentRestored(self, obj):
        # when files are shared it is essential to be able to change/set the shape file,
        # otherwise the file is hard to use
        # obj.setEditorMode('BitShape', 1)
        obj.setEditorMode('BitBody', 2)
        obj.setEditorMode('File', 1)
        obj.setEditorMode('Shape', 2)
        if not hasattr(obj, 'BitPropertyNames'):
            obj.addProperty('App::PropertyStringList', 'BitPropertyNames', 'Base', translate('PathToolBit', 'List of all properties inherited from the bit'))
            propNames = []
            for prop in obj.PropertiesList:
                if obj.getGroupOfProperty(prop) == 'Bit':
                    val = obj.getPropertyByName(prop)
                    typ = obj.getTypeIdOfProperty(prop)
                    dsc = obj.getDocumentationOfProperty(prop)

                    obj.removeProperty(prop)
                    obj.addProperty(typ, prop, PropertyGroupShape, dsc)

                    PathUtil.setProperty(obj, prop, val)
                    propNames.append(prop)
                elif obj.getGroupOfProperty(prop) == 'Attribute':
                    propNames.append(prop)
            obj.BitPropertyNames = propNames
        obj.setEditorMode('BitPropertyNames', 2)

        for prop in obj.BitPropertyNames:
            if obj.getGroupOfProperty(prop) == PropertyGroupShape:
                # properties in the Shape group can only be modified while the actual
                # shape is loaded, so we have to disable direct property editing
                obj.setEditorMode(prop, 1)
            else:
                # all other custom properties can and should be edited directly in the
                # property editor widget, not much value in re-implementing that
                obj.setEditorMode(prop, 0)

    def onChanged(self, obj, prop):
        PathLog.track(obj.Label, prop)
        if prop == 'BitShape' and 'Restore' not in obj.State:
            self._setupBitShape(obj)

    def onDelete(self, obj, arg2=None):
        PathLog.track(obj.Label)
        self.unloadBitBody(obj)
        obj.Document.removeObject(obj.Name)

    def _updateBitShape(self, obj, properties=None):
        if obj.BitBody is not None:
            for attributes in [o for o in obj.BitBody.Group if hasattr(o, 'Proxy') and hasattr(o.Proxy, 'getCustomProperties')]:
                for prop in attributes.Proxy.getCustomProperties():
                    # the property might not exist in our local object (new attribute in shape)
                    # for such attributes we just keep the default
                    if hasattr(obj, prop):
                        setattr(attributes, prop, obj.getPropertyByName(prop))
                    else:
                        # if the template shape has a new attribute defined we should add that
                        # to the local object
                        self._setupProperty(obj, prop, attributes)
                        propNames = obj.BitPropertyNames
                        propNames.append(prop)
                        obj.BitPropertyNames = propNames
            self._copyBitShape(obj)

    def _copyBitShape(self, obj):
        obj.Document.recompute()
        if obj.BitBody and obj.BitBody.Shape:
            obj.Shape = obj.BitBody.Shape
        else:
            obj.Shape = Part.Shape()

    def _loadBitBody(self, obj, path=None):
        PathLog.track(obj.Label, path)
        p = path if path else obj.BitShape
        docOpened = False
        doc = None
        for d in FreeCAD.listDocuments():
            if FreeCAD.getDocument(d).FileName == p:
                doc = FreeCAD.getDocument(d)
                break
        if doc is None:
            p = findToolShape(p, path if path else obj.File)
            if not path and p != obj.BitShape:
                obj.BitShape = p
            PathLog.debug("ToolBit {} using shape file: {}".format(obj.Label, p))
            doc = FreeCAD.openDocument(p, True)
            obj.ShapeName = doc.Name
            docOpened = True
        else:
            PathLog.debug("ToolBit {} already open: {}".format(obj.Label, doc))
        return (doc, docOpened)

    def _removeBitBody(self, obj):
        if obj.BitBody:
            obj.BitBody.removeObjectsFromDocument()
            obj.Document.removeObject(obj.BitBody.Name)
            obj.BitBody = None

    def _deleteBitSetup(self, obj):
        PathLog.track(obj.Label)
        self._removeBitBody(obj)
        self._copyBitShape(obj)
        for prop in obj.BitPropertyNames:
            obj.removeProperty(prop)

    def loadBitBody(self, obj, force=False):
        if force or not obj.BitBody:
            activeDoc = FreeCAD.ActiveDocument
            if force:
                self._removeBitBody(obj)
            (doc, opened) = self._loadBitBody(obj)
            obj.BitBody = obj.Document.copyObject(doc.RootObjects[0], True)
            if opened:
                FreeCAD.setActiveDocument(activeDoc.Name)
                FreeCAD.closeDocument(doc.Name)
            self._updateBitShape(obj)

    def unloadBitBody(self, obj):
        self._removeBitBody(obj)

    def _setupProperty(self, obj, prop, orig):
        # extract property parameters and values so it can be copied
        val = orig.getPropertyByName(prop)
        typ = orig.getTypeIdOfProperty(prop)
        grp = orig.getGroupOfProperty(prop)
        dsc = orig.getDocumentationOfProperty(prop)

        obj.addProperty(typ, prop, grp, dsc)
        if 'App::PropertyEnumeration' == typ:
            setattr(obj, prop, orig.getEnumerationsOfProperty(prop))

        obj.setEditorMode(prop, 1)
        PathUtil.setProperty(obj, prop, val)

    def _setupBitShape(self, obj, path=None):
        PathLog.track(obj.Label)

        activeDoc = FreeCAD.ActiveDocument
        (doc, docOpened) = self._loadBitBody(obj, path)

        obj.Label = doc.RootObjects[0].Label
        self._deleteBitSetup(obj)
        bitBody = obj.Document.copyObject(doc.RootObjects[0], True)

        docName = doc.Name
        if docOpened:
            FreeCAD.setActiveDocument(activeDoc.Name)
            FreeCAD.closeDocument(doc.Name)

        if bitBody.ViewObject:
            bitBody.ViewObject.Visibility = False

        PathLog.debug("bitBody.{} ({}): {}".format(bitBody.Label, bitBody.Name, type(bitBody)))

        propNames = []
        for attributes in [o for o in bitBody.Group if PathPropertyBag.IsPropertyBag(o)]:
            PathLog.debug("Process properties from {}".format(attributes.Label))
            for prop in attributes.Proxy.getCustomProperties():
                self._setupProperty(obj, prop, attributes)
                propNames.append(prop)
        if not propNames:
            PathLog.error(translate('PathToolBit', 'Did not find a PropertyBag in {} - not a ToolBit shape?'.format(docName)))

        # has to happen last because it could trigger op.execute evaluations
        obj.BitPropertyNames = propNames
        obj.BitBody = bitBody
        self._copyBitShape(obj)

    def toolShapeProperties(self, obj):
        '''toolShapeProperties(obj) ... return all properties defining it's shape'''
        return sorted([prop for prop in obj.BitPropertyNames if obj.getGroupOfProperty(prop) == PropertyGroupShape])

    def toolAdditionalProperties(self, obj):
        '''toolShapeProperties(obj) ... return all properties unrelated to it's shape'''
        return sorted([prop for prop in obj.BitPropertyNames if obj.getGroupOfProperty(prop) != PropertyGroupShape])

    def toolGroupsAndProperties(self, obj, includeShape=True):
        '''toolGroupsAndProperties(obj) ... returns a dictionary of group names with a list of property names.'''
        category = {}
        for prop in obj.BitPropertyNames:
            group = obj.getGroupOfProperty(prop)
            if includeShape or group != PropertyGroupShape:
                properties  = category.get(group, [])
                properties.append(prop)
                category[group] = properties
        return category

    def getBitThumbnail(self, obj):
        if obj.BitShape:
            path = findToolShape(obj.BitShape)
            if path:
                with open(path, 'rb') as fd:
                    try:
                        zf = zipfile.ZipFile(fd)
                        pf = zf.open('thumbnails/Thumbnail.png', 'r')
                        data = pf.read()
                        pf.close()
                        return data
                    except KeyError:
                        pass
        return None

    def saveToFile(self, obj, path, setFile=True):
        PathLog.track(path)
        try:
            with open(path, 'w') as fp:
                json.dump(self.templateAttrs(obj), fp, indent='  ')
            if setFile:
                obj.File = path
            return True
        except (OSError, IOError) as e:
            PathLog.error("Could not save tool {} to {} ({})".format(obj.Label, path, e))
            raise

    def templateAttrs(self, obj):
        attrs = {}
        attrs['version'] = 2  # Path.Tool is version 1
        attrs['name'] = obj.Label
        if PathPreferences.toolsStoreAbsolutePaths():
            attrs['shape'] = obj.BitShape
        else:
            # attrs['shape'] = findRelativePathShape(obj.BitShape)
            # Extract the name of the shape file
            __, filShp = os.path.split(obj.BitShape)  #  __ is an ignored placeholder acknowledged by LGTM
            attrs['shape'] = str(filShp)
        params = {}
        for name in obj.BitPropertyNames:
            params[name] = PathUtil.getPropertyValueString(obj, name)
        attrs['parameter'] = params
        params = {}
        attrs['attribute'] = params
        return attrs


def Declaration(path):
    PathLog.track(path)
    with open(path, 'r') as fp:
        return json.load(fp)


class ToolBitFactory(object):

    def CreateFromAttrs(self, attrs, name='ToolBit', path=None):
        PathLog.track(attrs, path)
        obj = Factory.Create(name, attrs['shape'], path)
        obj.Label = attrs['name']
        params = attrs['parameter']
        for prop in params:
            PathUtil.setProperty(obj, prop, params[prop])
        obj.Proxy._updateBitShape(obj)
        obj.Proxy.unloadBitBody(obj)
        return obj

    def CreateFrom(self, path, name='ToolBit'):
        PathLog.track(name, path)
        try:
            data = Declaration(path)
            bit = Factory.CreateFromAttrs(data, name, path)
            return bit
        except (OSError, IOError) as e:
            PathLog.error("%s not a valid tool file (%s)" % (path, e))
            raise

    def Create(self, name='ToolBit', shapeFile=None, path=None):
        PathLog.track(name, shapeFile, path)
        obj = FreeCAD.ActiveDocument.addObject('Part::FeaturePython', name)
        obj.Proxy = ToolBit(obj, shapeFile, path)
        return obj


Factory = ToolBitFactory()



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2018 Benjamin Alterauge (ageeye)                        *
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the PointArray object.

To Do
-----
This object currently uses the points inside a `Draft Block`,
a `Part::Compound`, or a `Sketch`. To make this more general,
this object could be augmented to extract a list of points
from the vertices of any object with a `Part::TopoShape` (2D or 3D).

See the `get_point_list` function for more information.
"""
## @package pointarray
# \ingroup draftobjects
# \brief Provides the object code for the PointArray object.

import lazy_loader.lazy_loader as lz
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import draftutils.utils as utils

from draftutils.messages import _wrn, _err
from draftutils.translate import translate
from draftobjects.draftlink import DraftLink

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftobjects
# @{


class PointArray(DraftLink):
    """The Draft Point Array object."""

    def __init__(self, obj):
        super(PointArray, self).__init__(obj, "PointArray")

    def attach(self, obj):
        """Set up the properties when the object is attached."""
        self.set_properties(obj)
        super(PointArray, self).attach(obj)

    def linkSetup(self, obj):
        """Set up the object as a link object."""
        super(PointArray, self).linkSetup(obj)
        obj.configLinkProperty(ElementCount='Count')

    def set_properties(self, obj):
        """Set properties only if they don't exist."""
        properties = obj.PropertiesList

        if "Base" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Base object that will be duplicated")
            obj.addProperty("App::PropertyLink",
                            "Base",
                            "Objects",
                            _tip)
            obj.Base = None

        if "PointObject" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Object containing points used to distribute the base object, for example, a sketch or a Part compound.\nThe sketch or compound must contain at least one explicit point or vertex object.")
            obj.addProperty("App::PropertyLink",
                            "PointObject",
                            "Objects",
                            _tip)
            obj.PointObject = None

        if "Count" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Total number of elements in the array.\nThis property is read-only, as the number depends on the points contained within 'Point Object'.")
            obj.addProperty("App::PropertyInteger",
                            "Count",
                            "Objects",
                            _tip)
            obj.Count = 0
            obj.setEditorMode("Count", 1)  # Read only

        if "ExtraPlacement" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Additional placement, shift and rotation, that will be applied to each copy")
            obj.addProperty("App::PropertyPlacement",
                            "ExtraPlacement",
                            "Objects",
                            _tip)
            obj.ExtraPlacement = App.Placement()

        if self.use_link and "ExpandArray" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property","Show the individual array elements (only for Link arrays)")
            obj.addProperty("App::PropertyBool",
                            "ExpandArray",
                            "Objects",
                            _tip)
            obj.setPropertyStatus('Shape', 'Transient')

    def execute(self, obj):
        """Run when the object is created or recomputed."""

        pt_list, count = get_point_list(obj.PointObject)
        pls = build_placements(obj.Base, pt_list, obj.ExtraPlacement)

        return super(PointArray, self).buildShape(obj, obj.Placement, pls)

    def onDocumentRestored(self, obj):
        """Execute code when the document is restored.

        Add properties that don't exist and migrate old properties.
        """
        # If the ExtraPlacement property has never been added before
        # it will add it first, and set it to the base object's position
        # in order to produce the same displacement as before.
        # Then all the other properties will be processed.
        properties = obj.PropertiesList

        if "ExtraPlacement" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Additional placement, shift and rotation, that will be applied to each copy")
            obj.addProperty("App::PropertyPlacement",
                            "ExtraPlacement",
                            "Objects",
                            _tip)
            obj.ExtraPlacement.Base = obj.Base.Placement.Base
            _wrn("v0.19, " + obj.Label + ", " + translate("draft","added property 'ExtraPlacement'"))

        self.set_properties(obj)
        self.migrate_properties_0v19(obj)
        return super(PointArray, self).onDocumentRestored(obj)

    def migrate_properties_0v19(self, obj):
        """Migrate properties."""
        # If the old name still exists, migrate it to the new
        # name and delete the old property
        properties = obj.PropertiesList

        if "PointList" in properties:
            obj.PointObject = obj.PointList
            obj.removeProperty("PointList")
            _info = "'PointList' property will be migrated to 'PointObject'"
            _wrn("v0.19, " + obj.Label + ", " + translate("draft","added property 'ExtraPlacement'"))


def get_point_list(point_object):
    """Extract a list of points from a point object.

    Parameters
    ----------
    point_object: Part::Feature
        Either a `Draft Block`, a `Part::Compound`,
        or a `Sketcher::SketchObject` containing points.

    Returns
    -------
    list, int
        A list of points that have `X`, `Y`, `Z` coordinates;
        the second element is the number of elements.
        If the list is empty, the second element is zero.

    To Do
    -----
    - This function currently extracts the points inside a `Draft Block`,
      a `Part::Compound`, or a `Sketch`. To make this more general,
      this function could be augmented to extract a list of points
      from the vertices of any object with a `Part::TopoShape` (2D or 3D).

    - If the input is a `Part::Compound`, it should handle all valid types
      of objects simultaneously, that is, `Draft Points`, `Part::Vertexes`,
      `Sketches` with points, and possibly any other object with
      a `Part::TopoShape` as in the previous point.
      It should recursively call itself to extract
      points contained in nested compounds.
    """
    # If its a clone, extract the real object
    while utils.get_type(point_object) == 'Clone':
        point_object = point_object.Objects[0]

    # If the point object doesn't have actual points
    # the point list will remain empty
    pt_list = list()

    if hasattr(point_object, 'Geometry'):
        # Intended for a Sketcher::SketchObject, which has this property
        place = point_object.Placement
        for geo in point_object.Geometry:
            # It must contain at least one Part::GeomPoint.
            if (hasattr(geo, 'X')
                    and hasattr(geo, 'Y') and hasattr(geo, 'Z')):
                point = geo.copy()
                point.translate(place.Base)
                point.rotate(place)
                pt_list.append(point)

        count = len(pt_list)
        return pt_list, count

    obj_list = list()
    if hasattr(point_object, 'Links'):
        # Intended for a Part::Compound, which has this property
        obj_list = point_object.Links
    elif hasattr(point_object, 'Components'):
        # Intended for a Draft Block, which has this property
        obj_list = point_object.Components

    # These compounds should have at least one discrete point object
    # like a Draft Point or a Part::Vertex
    for _obj in obj_list:
        if hasattr(_obj, 'X') and hasattr(_obj, 'Y') and hasattr(_obj, 'Z'):
            pt_list.append(_obj)

    count = len(pt_list)
    return pt_list, count

def build_placements(base_object, pt_list=None, placement=App.Placement()):
    """Build a placements from the base object and list of points.

    Returns
    -------
    list(App.Placement)
    """
    if not pt_list:
        _err(translate("Draft",
                       "Point object doesn't have a discrete point, "
                       "it cannot be used for an array."))
        return []

    pls = list()

    for point in pt_list:
        new_pla = base_object.Placement.copy()
        original_rotation = new_pla.Rotation

        # Reset the position of the copy, and combine the original rotation
        # with the provided rotation. Two rotations (quaternions)
        # are combined by multiplying them.
        new_pla.Base = placement.Base
        new_pla.Rotation = original_rotation * placement.Rotation

        if point.TypeId == "Part::Vertex":
            # For this object the final position is the value of the Placement
            # plus the value of the X, Y, Z properties
            place = App.Vector(point.X,
                               point.Y,
                               point.Z) + point.Placement.Base

        elif hasattr(point, 'Placement'):
            # If the point object has a placement (Draft Point), use it
            # to displace the copy of the shape
            place = point.Placement.Base

            # The following old code doesn't make much sense because it uses
            # the rotation value of the auxiliary point.
            # Even if the point does have a rotation property as part of its
            # placement, rotating a point by itself is a strange workflow.
            # We want to use the position of the point but not its rotation,
            # which will probably be zero anyway.

            # Old code:
            # place = point.Placement
            # new_shape.rotate(place.Base,
            #                  place.Rotation.Axis,
            #                  math.degrees(place.Rotation.Angle))
        else:
            # In other cases (Sketch with points)
            # translate by the X, Y, Z coordinates
            place = App.Vector(point.X, point.Y, point.Z)

        new_pla.translate(place)

        pls.append(new_pla)

    return pls

def build_copies(base_object, pt_list=None, placement=App.Placement()):
    """Build a compound of copies from the base object and list of points.

    Returns
    -------
    Part::TopoShape
        The compound shape created by `Part.makeCompound`.
    """

    if not pt_list:
        _err(translate("Draft",
                       "Point object doesn't have a discrete point, "
                       "it cannot be used for an array."))
        shape = base_object.Shape.copy()
        return shape

    copies = list()

    for pla in build_copies(base_object, pt_list, placement):
        new_shape = base_object.Shape.copy()
        new_shape.Placement = pla

        copies.append(new_shape)

    shape = Part.makeCompound(copies)
    return shape


# Alias for compatibility with v0.18 and earlier
_PointArray = PointArray

## @}



# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for the Layer object."""
## @package view_layer
# \ingroup draftviewproviders
# \brief Provides the viewprovider code for the Layer object.

## \addtogroup draftviewproviders
# @{
import pivy.coin as coin
import PySide.QtCore as QtCore
import PySide.QtGui as QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui

from draftutils.messages import _msg
from draftutils.translate import translate
from draftobjects.layer import Layer


class ViewProviderLayer:
    """The viewprovider for the Layer object."""

    def __init__(self, vobj):
        self.Object = vobj.Object
        self.set_properties(vobj)

        vobj.Proxy = self

    def set_properties(self, vobj):
        """Set the properties only if they don't already exist."""
        properties = vobj.PropertiesList
        self.set_override_options(vobj, properties)
        self.set_visual_properties(vobj, properties)

    def set_override_options(self, vobj, properties):
        """Set property options only if they don't already exist."""
        if "OverrideLineColorChildren" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "If it is true, the objects contained "
                                     "within this layer will adopt "
                                     "the line color of the layer")
            vobj.addProperty("App::PropertyBool",
                             "OverrideLineColorChildren",
                             "Layer",
                             _tip)
            vobj.OverrideLineColorChildren = True

        if "OverrideShapeColorChildren" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "If it is true, the objects contained "
                                     "within this layer will adopt "
                                     "the line color of the layer")
            vobj.addProperty("App::PropertyBool",
                             "OverrideShapeColorChildren",
                             "Layer",
                             _tip)
            vobj.OverrideShapeColorChildren = True

        if "UsePrintColor" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "If it is true, the print color "
                                     "will be used when objects in this "
                                     "layer are placed on a TechDraw page")
            vobj.addProperty("App::PropertyBool",
                             "UsePrintColor",
                             "Print",
                             _tip)


    def set_visual_properties(self, vobj, properties):
        """Set visual properties only if they don't already exist."""
        view_group = App.ParamGet("User parameter:BaseApp/Preferences/View")

        if "LineColor" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The line color of the objects "
                                     "contained within this layer")
            vobj.addProperty("App::PropertyColor",
                             "LineColor",
                             "Layer",
                             _tip)

            c = view_group.GetUnsigned("DefaultShapeLineColor", 255)
            vobj.LineColor = (((c >> 24) & 0xFF) / 255,
                              ((c >> 16) & 0xFF) / 255,
                              ((c >> 8) & 0xFF) / 255)

        if "ShapeColor" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The shape color of the objects "
                                     "contained within this layer")
            vobj.addProperty("App::PropertyColor",
                             "ShapeColor",
                             "Layer",
                             _tip)

            c = view_group.GetUnsigned("DefaultShapeColor", 4294967295)
            vobj.ShapeColor = (((c >> 24) & 0xFF) / 255,
                               ((c >> 16) & 0xFF) / 255,
                               ((c >> 8) & 0xFF) / 255)

        if "LineWidth" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The line width of the objects contained "
                                     "within this layer")
            vobj.addProperty("App::PropertyFloat",
                             "LineWidth",
                             "Layer",
                             _tip)

            width = view_group.GetInt("DefaultShapeLineWidth", 2)
            vobj.LineWidth = width

        if "DrawStyle" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The draw style of the objects contained "
                                     "within this layer")
            vobj.addProperty("App::PropertyEnumeration",
                             "DrawStyle",
                             "Layer",
                             _tip)
            vobj.DrawStyle = ["Solid", "Dashed", "Dotted", "Dashdot"]
            vobj.DrawStyle = "Solid"

        if "Transparency" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The transparency of the objects "
                                     "contained within this layer")
            vobj.addProperty("App::PropertyInteger",
                             "Transparency",
                             "Layer",
                             _tip)
            vobj.Transparency = 0

        if "LinePrintColor" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The line color of the objects "
                                     "contained within this layer, "
                                     "when used on a TechDraw page")
            vobj.addProperty("App::PropertyColor",
                             "LinePrintColor",
                             "Print",
                             _tip)

    def getIcon(self):
        """Return the path to the icon used by the viewprovider.

        Normally it returns the basic Layer icon, but if `icondata` exists
        it is the modified icon with the line and shape colors of the layer.
        """
        if hasattr(self, "icondata"):
            return self.icondata
        return ":/icons/Draft_Layer.svg"

    def attach(self, vobj):
        """Set up the scene sub-graph of the viewprovider."""
        self.Object = vobj.Object
        sep = coin.SoGroup()
        vobj.addDisplayMode(sep, "Default")

    def claimChildren(self):
        """Return objects that will be placed under it in the tree view.

        These are the elements of the `Group` property of the Proxy object.
        """
        if hasattr(self, "Object") and hasattr(self.Object, "Group"):
            return self.Object.Group

    def getDisplayModes(self, vobj):
        """Return the display modes that this viewprovider supports."""
        return ["Default"]

    def getDefaultDisplayMode(self):
        """Return the default display mode."""
        return "Default"

    def setDisplayMode(self, mode):
        """Return the saved display mode."""
        return mode

    def __getstate__(self):
        """Return a tuple of objects to save or None."""
        return None

    def __setstate__(self, state):
        """Set the internal properties from the restored state."""
        return None

    def updateData(self, obj, prop):
        """Execute when a property from the Proxy class is changed."""
        if prop == "Group":
            for _prop in ("LineColor", "ShapeColor", "LineWidth",
                          "DrawStyle", "Transparency", "Visibility"):
                self.onChanged(obj.ViewObject, _prop)

    def change_view_properties(self, vobj, prop):
        """Iterate over the contents and change the properties."""
        obj = vobj.Object

        # Return if the property does not exist
        if not hasattr(vobj, prop):
            return

        for target_obj in obj.Group:
            target_vobj = target_obj.ViewObject

            # If the override properties are not set return without change
            if prop == "LineColor" and not vobj.OverrideLineColorChildren:
                return
            elif prop == "ShapeColor" and not vobj.OverrideShapeColorChildren:
                return

            # This checks that the property exists in the target object,
            # and then sets the target property accordingly
            if hasattr(target_vobj, prop):
                setattr(target_vobj, prop, getattr(vobj, prop))
            else:
                continue

            # Use the line color for the text color if it exists
            if prop == "LineColor":
                if hasattr(target_vobj, "TextColor"):
                    target_vobj.TextColor = vobj.LineColor
                if hasattr(target_vobj, "FontColor"):
                    target_vobj.FontColor = vobj.LineColor

    def onChanged(self, vobj, prop):
        """Execute when a view property is changed."""
        if prop in ("LineColor", "ShapeColor", "LineWidth",
                    "DrawStyle", "Transparency", "Visibility"):
            self.change_view_properties(vobj, prop)

        if (prop in ("LineColor", "ShapeColor")
                and hasattr(vobj, "LineColor")
                and hasattr(vobj, "ShapeColor")):
            # This doesn't do anything to the objects inside the layer,
            # it just uses the defined Line and Shape colors
            # to paint the layer icon accordingly in the tree view
            l_color = vobj.LineColor
            s_color = vobj.ShapeColor

            l_color = QtGui.QColor(int(l_color[0] * 255),
                                   int(l_color[1] * 255),
                                   int(l_color[2] * 255))
            s_color = QtGui.QColor(int(s_color[0] * 255),
                                   int(s_color[1] * 255),
                                   int(s_color[2] * 255))
            p1 = QtCore.QPointF(2, 17)
            p2 = QtCore.QPointF(13, 8)
            p3 = QtCore.QPointF(30, 15)
            p4 = QtCore.QPointF(20, 25)

            image = QtGui.QImage(32, 32, QtGui.QImage.Format_ARGB32)
            image.fill(QtCore.Qt.transparent)

            pt = QtGui.QPainter(image)
            pt.setBrush(QtGui.QBrush(s_color, QtCore.Qt.SolidPattern))
            pt.drawPolygon([p1, p2, p3, p4])
            pt.setPen(QtGui.QPen(l_color, 2,
                                 QtCore.Qt.SolidLine, QtCore.Qt.FlatCap))
            pt.drawPolygon([p1, p2, p3, p4])
            pt.end()

            byte_array = QtCore.QByteArray()
            buffer = QtCore.QBuffer(byte_array)
            buffer.open(QtCore.QIODevice.WriteOnly)
            image.save(buffer, "XPM")

            self.icondata = byte_array.data().decode("latin1")
            vobj.signalChangeIcon()

    def canDragObject(self, obj):
        """Return True to allow dragging one object from the Layer."""
        return True

    def canDragObjects(self):
        """Return True to allow dragging many objects from the Layer."""
        return True

    def dragObject(self, vobj, otherobj):
        """Remove the object that was dragged from the layer."""
        if hasattr(vobj.Object, "Group") and otherobj in vobj.Object.Group:
            group = vobj.Object.Group
            group.remove(otherobj)
            vobj.Object.Group = group
            App.ActiveDocument.recompute()

    def canDropObject(self, obj):
        """Return true to allow dropping one object.

        If the object being dropped is itself a `'Layer'`, return `False`
        to prevent dropping a layer inside a layer, at least for now.
        """
        if hasattr(obj, "Proxy") and isinstance(obj.Proxy, Layer):
            return False
        return True

    def canDropObjects(self):
        """Return true to allow dropping many objects."""
        return True

    def dropObject(self, vobj, otherobj):
        """Add object that was dropped into the Layer to the group.

        If the object being dropped is itself a `'Layer'`,
        return immediately to prevent dropping a layer inside a layer,
        at least for now.
        """
        if hasattr(otherobj, "Proxy") and isinstance(otherobj.Proxy, Layer):
            return

        obj = vobj.Object

        if hasattr(obj, "Group") and otherobj not in obj.Group:
            group = obj.Group
            group.append(otherobj)
            obj.Group = group

            # Remove from all other layers (not automatic)
            for parent in otherobj.InList:
                if (hasattr(parent, "Proxy")
                        and isinstance(parent.Proxy, Layer)
                        and otherobj in parent.Group
                        and parent != obj):
                    p_group = parent.Group
                    p_group.remove(otherobj)
                    parent.Group = p_group

            App.ActiveDocument.recompute()

    def setupContextMenu(self, vobj, menu):
        """Set up actions to perform in the context menu."""
        action1 = QtGui.QAction(QtGui.QIcon(":/icons/button_right.svg"),
                                translate("draft", "Activate this layer"),
                                menu)
        action1.triggered.connect(self.activate)
        menu.addAction(action1)

        action2 = QtGui.QAction(QtGui.QIcon(":/icons/Draft_SelectGroup.svg"),
                                translate("draft", "Select layer contents"),
                                menu)
        action2.triggered.connect(self.select_contents)
        menu.addAction(action2)

    def activate(self):
        """Activate the selected layer, it becomes the Autogroup."""
        if hasattr(self, "Object"):
            Gui.Selection.clearSelection()
            Gui.Selection.addSelection(self.Object)
            Gui.runCommand("Draft_AutoGroup")

    def select_contents(self):
        """Select the contents of the layer."""
        if hasattr(self, "Object"):
            Gui.Selection.clearSelection()
            for layer_obj in self.Object.Group:
                Gui.Selection.addSelection(layer_obj)


class ViewProviderLayerContainer:
    """The viewprovider for the LayerContainer object."""

    def __init__(self, vobj):
        self.Object = vobj.Object
        vobj.Proxy = self

    def getIcon(self):
        """Return the path to the icon used by the viewprovider."""
        return ":/icons/Draft_Layer.svg"

    def attach(self, vobj):
        """Set up the scene sub-graph of the viewprovider."""
        self.Object = vobj.Object

    def setupContextMenu(self, vobj, menu):
        """Set up actions to perform in the context menu."""
        action1 = QtGui.QAction(QtGui.QIcon(":/icons/Draft_Layer.svg"),
                                translate("Draft", "Merge layer duplicates"),
                                menu)
        action1.triggered.connect(self.merge_by_name)
        menu.addAction(action1)
        action2 = QtGui.QAction(QtGui.QIcon(":/icons/Draft_NewLayer.svg"),
                                translate("Draft", "Add new layer"),
                                menu)
        action2.triggered.connect(self.add_layer)
        menu.addAction(action2)

    def merge_by_name(self):
        """Merge the layers that have the same name."""
        if not hasattr(self, "Object") or not hasattr(self.Object, "Group"):
            return

        obj = self.Object

        layers = list()
        for iobj in obj.Group:
            if hasattr(iobj, "Proxy") and isinstance(iobj.Proxy, Layer):
                layers.append(iobj)

        to_delete = list()
        for layer in layers:
            # Test the last three characters of the layer's Label to see
            # if it's a number, like `'Layer017'`
            if (layer.Label[-1].isdigit()
                    and layer.Label[-2].isdigit()
                    and layer.Label[-3].isdigit()):
                # If the object inside the layer has the same Label
                # as the layer, save this object
                orig = None
                for ol in layer.OutList:
                    if ol.Label == layer.Label[:-3].strip():
                        orig = ol
                        break

                # Go into the objects that reference this layer object
                # and set the layer property with the previous `orig`
                # object found
                # Editor: when is this possible? Maybe if a layer is inside
                # another layer? Currently the code doesn't allow this
                # so maybe this was a previous behavior that was disabled
                # in `ViewProviderLayer`.
                if orig:
                    for par in layer.InList:
                        for prop in par.PropertiesList:
                            if getattr(par, prop) == layer:
                                _msg("Changed property '" + prop
                                     + "' of object " + par.Label
                                     + " from " + layer.Label
                                     + " to " + orig.Label)
                                setattr(par, prop, orig)
                    to_delete.append(layer)

        for layer in to_delete:
            if not layer.InList:
                _msg("Merging duplicate layer: " + layer.Label)
                App.ActiveDocument.removeObject(layer.Name)
            elif len(layer.InList) == 1:
                first = layer.InList[0]

                if first.isDerivedFrom("App::DocumentObjectGroup"):
                    _msg("Merging duplicate layer: " + layer.Label)
                    App.ActiveDocument.removeObject(layer.Name)
            else:
                _msg("InList not empty. "
                     "Unable to delete layer: " + layer.Label)

    def add_layer(self):
        """Creates a new layer"""
        import Draft
        Draft.make_layer()
        App.ActiveDocument.recompute()

    def __getstate__(self):
        """Return a tuple of objects to save or None."""
        return None

    def __setstate__(self, state):
        """Set the internal properties from the restored state."""
        return None


# Alias for compatibility with v0.18 and earlier
_ViewProviderVisGroup = ViewProviderLayer

## @}
